from unittest import TestCase


class GetOnlineTargets(TestCase):
    """
    tested in:
        pcs.lib.commands.test.sbd.test_enable_sbd
    """

class SendPcsdSslCertAndKey(TestCase):
    """
    tested in:
        pcs.lib.commands.test.cluster.test_setup
        pcs.lib.commands.test.test_pcsd
    """

class RemoveNodesFromCib(TestCase):
    """
    tested in:
        pcs.lib.commands.test.cluster.test_remove_nodes
            .RemoveNodesFailureFromCib
        pcs.lib.commands.test.cluster.test_remove_nodes
            .RemoveNodesSuccessMinimal
    """
